#!/usr/bin/env python3

import cgi, sys, os, datetime, re, json, yaml
from os import environ, pardir, path
from pymongo import MongoClient

from bycon import *

"""podmd

* <https://progenetix.org/beacon/datasets/>

podmd"""

################################################################################
################################################################################
################################################################################

def main():

    try:
        cohorts()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)

################################################################################

def collections():

    try:
        cohorts()
    except Exception:
        print_text_response(traceback.format_exc(), byc["env"], 302)
    
################################################################################

def cohorts():

    initialize_bycon_service(byc)
    run_beacon_init_stack(byc)

    _return_cohorts_response(byc)

    run_result_sets_beacon(byc)
    query_results_save_handovers(byc)
    check_computed_interval_frequency_delivery(byc)
    check_switch_to_count_response(byc)
    check_switch_to_boolean_response(byc)
    cgi_print_response( byc, 200 )

################################################################################

def _return_cohorts_response(byc):

    if not "cohort" in byc["response_entity_id"]:
        return

    mongo_client = MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))

    cohorts =  []
    c_id = byc.get("request_entity_path_id_value")
 
    if c_id is not None:
        byc["queries"].update( {"cohorts": { "id": c_id } } )
    else:
        byc["queries"].update( {"cohorts": { "collation_type": "pgxcohort" } } )

    query = byc["queries"].get("cohorts", {})
    
    for ds_id in byc[ "dataset_ids" ]:
        mongo_db = mongo_client[ ds_id ]        
        mongo_coll = mongo_db[ "collations" ]

        for cohort in mongo_coll.find( query ):
            cohorts.append(cohort)
            byc["service_response"]["response_summary"].update({"exists":True})

    cohorts = remap_cohorts(cohorts, byc)

    if byc["test_mode"] is True:
        ret_no = int(byc.get('test_mode_count', 5))
        cohorts = cohorts[:ret_no]

    byc["service_response"]["response"].pop("result_sets", None)
    byc["service_response"]["response"].update({"collections": cohorts})
    byc["service_response"]["response_summary"].update({"num_total_results":len(cohorts)})
    cgi_print_response( byc, 200 )

################################################################################
################################################################################

if __name__ == '__main__':
    main()



from pymongo import MongoClient
from os import environ

################################################################################

def retrieve_data(ds_id, byc):

    r_c = byc["response_entity"]["collection"]
    r_k = r_c+"_id"

    for r_t, r_d in byc["beacon_mappings"]["response_types"].items():
        # print(r_d["entity_type"], byc["response_entity_id"])

        if r_d["entity_type"] == byc["response_entity_id"]:
            r_k = r_d["h->o_access_key"]

    if "variants" in r_c:
        r_s_res = retrieve_variants(ds_id, byc)
        return r_s_res

    mongo_client = MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
    data_db = mongo_client[ ds_id ]
    data_coll = mongo_client[ ds_id ][ r_c ]

    ds_results = byc["dataset_results"][ds_id]

    if r_k in ds_results:
        r_s_res = []
        for d in data_coll.find({"_id":{"$in": ds_results[ r_k ]["target_values"] }}):
            r_s_res.append(d)

        return r_s_res

    return []

################################################################################

def retrieve_variants(ds_id, byc):

    ds_results = byc["dataset_results"][ds_id]

    if "all_variants_methods" in byc["service_config"]:
        if byc["method"] in byc["service_config"]["all_variants_methods"]:
            return False

    mongo_client = MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
    data_db = mongo_client[ ds_id ]
    v_coll = mongo_client[ ds_id ][ "variants" ]

    r_s_res = []

    if "variants._id" in ds_results:
        for v_id in ds_results["variants._id"]["target_values"]:
            v = v_coll.find_one({"_id":v_id})
            r_s_res.append(v)
        return r_s_res
    elif "variants.variant_internal_id" in ds_results:
        for v_id in ds_results["variants.variant_internal_id"]["target_values"]:
            vs = v_coll.find({"variant_internal_id":v_id})
            for v in vs:
                r_s_res.append(v)
        return r_s_res

    return False



import pymongo
from bson import SON

from bycon_helpers import days_from_iso8601duration, paginate_list, set_pagination_range
from filter_parsing import *
from variant_parsing import *


################################################################################

def generate_dataset_queries(byc, ds_id):

    byc.update({"queries": {}})
    __update_queries_from_path_id(byc)
    __update_queries_from_id_values(byc)
    __update_queries_from_cohorts_query(byc)
    __update_queries_from_variants(byc)
    __update_queries_from_geoquery(byc)
    __update_queries_from_hoid(byc)
    __update_queries_from_filters(byc, ds_id)
    __replace_queries_in_test_mode(byc)
    __purge_empty_queries(byc)

    # TODO: HOT FIX
    if "runs" in byc["queries"].keys():
        if "callsets" not in byc["queries"].keys():
            byc["queries"]["callsets"] = byc["queries"].pop("runs")


################################################################################

def __purge_empty_queries(byc):
    empties = []
    for k, v in byc["queries"].items():
        if not v:
            empties.append(k)
    for e_k in empties:
        byc["queries"].pop(e_k, None)


################################################################################

def __replace_queries_in_test_mode(byc):
    if byc["test_mode"] is not True:
        return

    try:
        collname = byc["response_entity"]["collection"]
    except Exception:
        return

    ret_no = int(byc.get('test_mode_count', 5))

    ds_id = byc["dataset_ids"][0]
    mongo_client = pymongo.MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
    data_db = mongo_client[ds_id]
    data_collnames = data_db.list_collection_names()

    if collname not in data_collnames:
        return

    data_coll = mongo_client[ds_id][collname]
    rs = list(data_coll.aggregate([{"$sample": {"size": ret_no}}]))

    _ids = []
    for r in rs:
        _ids.append(r["_id"])

    byc.update({
        "queries": {collname: {"_id": {"$in": _ids}}},
        "empty_query_all_count": data_coll.estimated_document_count()
    })


################################################################################

def __update_queries_from_path_id(byc):
    b_mps = byc["beacon_mappings"]

    if "service" in byc["request_path_root"]:
        b_mps = byc.get("services_mappings", byc["beacon_mappings"])

    if not byc["request_entity_id"]:
        return

    r_e_id = byc["request_entity_id"]
    p_id_v = byc["request_entity_path_id_value"]

    if not byc["request_entity_path_id_value"]:
        return

    if r_e_id not in b_mps["response_types"]:
        return

    collname = b_mps["response_types"][r_e_id]["collection"]

    if not collname:
        return

    byc["queries"].update({collname: {"id": p_id_v}})


################################################################################

def __update_queries_from_cohorts_query(byc):
    if "cohorts" not in byc["queries"]:
        return

    if "cohort" in byc["response_entity_id"]:
        return

    c_q = byc["queries"]["cohorts"]

    query = {}
    if "id" in c_q:
        query = {"cohorts.id": c_q["id"]}

    byc["queries"].pop("cohorts", None)

    __update_query_for_scope(byc, query, "biosamples")


################################################################################

def __update_queries_from_id_values(byc):
    id_f_v = byc["beacon_mappings"]["id_queryscope_mappings"]
    f_d = byc["form_data"]

    this_id_k = byc["response_entity_id"] + "_ids"

    if "ids" in f_d:
        if this_id_k not in f_d:
            f_d.update({this_id_k: f_d["ids"]})

    for id_k, id_s in id_f_v.items():
        q = False
        if id_k in f_d:
            id_v = f_d[id_k]
            if len(id_v) > 1:
                q = {"id": {"$in": id_v}}
            elif len(id_v) == 1:
                q = {"id": id_v[0]}
        if q is not False:
            __update_query_for_scope(byc, q, id_s, "AND")


################################################################################

def __update_queries_from_hoid(byc):
    if "accessid" in byc["form_data"]:

        accessid = byc["form_data"]["accessid"]
        ho_client = pymongo.MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
        ho_db = ho_client[byc["config"]["housekeeping_db"]]
        ho_coll = ho_db[byc["config"]["handover_coll"]]
        h_o = ho_coll.find_one({"id": accessid})

        # accessid overrides ... ?
        if h_o:
            t_k = h_o["target_key"]
            t_v = h_o["target_values"]
            c_n = h_o["target_collection"]
            t_c = h_o["target_count"]

            byc.update({"original_queries": h_o.get("original_queries", None)})

            set_pagination_range(t_c, byc)
            t_v = paginate_list(t_v, byc)
            h_o_q = {t_k: {'$in': t_v}}
            if c_n in byc["queries"]:
                byc["queries"].update({c_n: {'$and': [h_o_q, byc["queries"][c_n]]}})
            else:
                byc["queries"].update({c_n: h_o_q})


################################################################################

def __update_queries_from_filters(byc, ds_id="progenetix"):
    """The new version assumes that dataset_id, scope (collection) and field are
    stored in the collation entries. Only filters with exact match to an entry
    in the lookup "collations" collection will be evaluated.
    While the Beacon v2 protocol assumes a logical `AND` between filters, bycon
    has a slightly differing approach:
    * filters against the same field (in the same collection) are treated as
    logical `OR` since this seems, well, logical (since otherwise the query would
    fail or result in incomplete matches in the case of terms from the same term
    tree); and also allows the use of the same query object for hierarchical
    (`child_terms`) query expansion
    * the `bycon` API allows to pass a `filterLogic` parameter with either a
    global `AND` (default value) or `OR`

    CAVE: Filters are assumed to be id values in the collation collection
    OR have a `collationed` flag in the filter_definitions set to `false`.

    TODO: The `excluded` query option currently will **not** positively match
    entries with an `excluded` flag (this is a specific Phenopackets construct
    for e.g. defining patients which have one phenotype _but not_ another one).
    """

    f_defs = byc["filter_definitions"]
    f_lists = {}

    logic = byc["filter_flags"]["logic"]
    f_desc = byc["filter_flags"]["descendants"]
    # precision = byc[ "filter_flags" ][ "precision" ]

    mongo_client = pymongo.MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
    coll_coll = mongo_client[ds_id]["collations"]

    filters = byc.get("filters", [])

    f_infos = {}

    for f in filters:

        f_val = f["id"]
        f_neg = f.get("excluded", False)
        if re.compile(r'^!').match(f_val):
            f_neg = True
            f_val = re.sub(r'^!', '', f_val)

        f_desc = f.get("includeDescendantTerms", f_desc)
        f_scope = f.get("scope", False)

        f_info = coll_coll.find_one({"id": f_val})

        if f_info is None:

            for f_d in f_defs.values():
                f_re = re.compile(f_d["pattern"])
                if f_re.match(f_val):
                    f_info = {
                        "id": f_val,
                        "scope": f_d["scope"],
                        "db_key": f_d["db_key"],
                        "type": f_d.get("type", "ontology"),
                        "format": f_d.get("format", "___none___"),
                        "child_terms": [f_val]
                    }
                    f_desc = False
                    if f_d["collationed"] is True:
                        ftw = f'Filter `{f_val}` matches a `{f_d["scope"]}` pattern but is not in the list of existing `filtering_terms` for {ds_id}'
                        response_add_filter_warnings(byc, ftw)                        

        if f_info is None:
            f_info = {
                "id": f_val,
                "scope": "biosamples",
                "type": "___undefined___",
                "db_key": "___undefined___",
                "child_terms": [f_val]
            }

        if f_neg is True:
            f_info.update({"is_negated": True})

        f_field = f_info.get("db_key", "id")

        if f_scope is False:
            f_scope = f_info["scope"]

        if f_scope not in byc["config"].get("queried_collections", []):
            fsw = f'The scope `{f_scope}` for filter `{f_val}` is not in the queried models'
            response_add_filter_warnings(byc, fsw)

        if f_scope not in f_lists.keys():
            f_lists.update({f_scope: {}})
        if f_scope not in f_infos.keys():
            f_infos.update({f_scope: {}})

        if f_field not in f_lists[f_scope].keys():
            f_lists[f_scope].update({f_field: []})
        if f_field not in f_infos[f_scope].keys():
            f_infos[f_scope].update({f_field: f_info})

        # TODO: move somewhere; this is just for the age prototype
        if "alphanumeric" in f_info.get("type", "ontology"):
            f_class, comp, val = re.match(r'^(\w+):([<>=]+?)(\w[\w.]+?)$', f_info["id"]).group(1, 2, 3)
            if "iso8601duration" in f_info.get("format", "___none___"):
                val = days_from_iso8601duration(val)
            f_lists[f_scope][f_field].append(__mongo_comparator_query(comp, val))

        elif f_desc is True:
            if f_neg is True:
                f_lists[f_scope][f_field].append({'$nin': f_info["child_terms"]})
            else:
                f_lists[f_scope][f_field].extend(f_info["child_terms"])
        else:
            if f_neg is True:
                f_lists[f_scope][f_field].append({'$nin': [f_info["id"]]})
            else:
                f_lists[f_scope][f_field].append(f_info["id"])

    # creating the queries & combining w/ possible existing ones
    for f_scope in f_lists.keys():
        f_s_l = []
        for f_field, f_query_vals in f_lists[f_scope].items():
            if len(f_query_vals) == 1:
                f_s_l.append({f_field: f_query_vals[0]})
            else:
                if "alphanumeric" in f_infos[f_scope][f_field].get("type", "ontology"):
                    q_l = []
                    for a_q_v in f_query_vals:
                        q_l.append({f_field: a_q_v})
                    f_s_l.append({"$and": q_l})
                else:
                    f_s_l.append({f_field: {"$in": f_query_vals}})

        if f_scope in byc["queries"]:
            f_s_l.append(byc["queries"][f_scope])

        if len(f_s_l) == 1:
            byc["queries"].update({f_scope: f_s_l[0]})
        elif len(f_s_l) > 1:
            byc["queries"].update({f_scope: {logic: f_s_l}})


################################################################################

def __update_query_for_scope(byc, query, scope, bool_mode="AND"):
    logic = boolean_to_mongo_logic(bool_mode)

    if scope not in byc["queries"]:
        byc["queries"][scope] = query
    else:
        byc["queries"][scope] = {logic: [byc["queries"][scope], query]}


################################################################################

def __update_queries_from_geoquery(byc):
    geo_q, geo_pars = geo_query(byc)

    if not geo_q:
        return

    __update_query_for_scope(byc, geo_q, "biosamples", bool_mode="AND")


################################################################################

def __mongo_comparator_query(comparator, value):

    mongo_comps = {
        ">": '$gt',
        ">=": '$gte',
        "<": '$lt',
        "<=": '$lte',
        "=": '$eq'
    }

    c = mongo_comps.get(comparator, '$eq')

    return {c: value}

################################################################################


def __update_queries_from_variants(byc):
    if "variant_request_type" not in byc:
        return

    if byc["variant_request_type"] not in byc["variant_parameters"]["request_types"].keys():
        if "variants" not in byc["queries"]:
            return

    if "variantTypeRequest" in byc["variant_request_type"]:
        create_variantTypeRequest_query(byc)
    elif "variantIdRequest" in byc["variant_request_type"]:
        create_variantIdRequest_query(byc)
    elif "variantCNVrequest" in byc["variant_request_type"]:
        create_variantCNVrequest_query(byc)
    elif "variantAlleleRequest" in byc["variant_request_type"]:
        create_variantAlleleRequest_query(byc)
    elif "variantRangeRequest" in byc["variant_request_type"]:
        create_variantRangeRequest_query(byc)
    elif "geneVariantRequest" in byc["variant_request_type"]:
        create_geneVariantRequest_query(byc)


################################################################################
################################################################################
################################################################################

def geo_query(byc):
    geo_q = {}
    geo_pars = {}

    if "geoloc_definitions" not in byc:
        return geo_q, geo_pars

    g_p_defs = byc["geoloc_definitions"]["parameters"]
    g_p_rts = byc["geoloc_definitions"]["request_types"]
    geo_root = byc["geoloc_definitions"]["geo_root"]

    req_type = ""
    # TODO: Make this modular & fix the one_of interpretation to really only 1
    for rt in g_p_rts:
        g_p = {}
        min_p_no = 1
        mat_p_no = 0
        if "all_of" in g_p_rts[rt]:
            g_q_k = g_p_rts[rt]["all_of"]
            min_p_no = len(g_q_k)
        elif "one_of" in g_p_rts[rt]:
            g_q_k = g_p_rts[rt]["one_of"]
        else:
            continue

        # print(rt)
        # print(byc["form_data"]["filters"])
        all_p = g_p_rts[rt].get("any_of", []) + g_q_k

        for g_k in g_p_defs.keys():

            if g_k not in all_p:
                continue

            g_default = None
            if "default" in g_p_defs[g_k]:
                g_default = g_p_defs[g_k]["default"]

            # TODO: This is an ISO lower hack ...

            if g_k.lower() in byc["form_data"]:
                g_v = byc["form_data"][g_k.lower()]
            else:
                g_v = g_default
            if g_v is None:
                continue
            if not re.compile(g_p_defs[g_k]["pattern"]).match(str(g_v)):
                continue
            if "float" in g_p_defs[g_k]["type"]:
                g_p[g_k] = float(g_v)
            else:
                g_p[g_k] = g_v

            if g_k in g_q_k:
                mat_p_no += 1

        if mat_p_no < min_p_no:
            continue

        req_type = rt
        geo_pars = g_p

    if "city" in req_type:
        geo_q = return_geo_city_query(geo_root, geo_pars)
        return geo_q, geo_pars

    if "id" in req_type:
        geo_q = {"id": re.compile(geo_pars["id"], re.IGNORECASE)}
        return geo_q, geo_pars

    if "ISO3166alpha2" in req_type:
        geo_q = {"provenance.geo_location.properties.ISO3166alpha2": byc["form_data"]["iso3166alpha2"].upper()}
        return geo_q, geo_pars

    if "geoquery" in req_type:
        geoq_l = [return_geo_longlat_query(geo_root, geo_pars)]
        for g_k in g_p_rts["geoquery"]["any_of"]:
            if g_k in geo_pars.keys():
                g_v = geo_pars[g_k]
                if len(geo_root) > 0:
                    geopar = ".".join([geo_root, "properties", g_k])
                else:
                    geopar = ".".join(["properties", g_k])
                geoq_l.append({geopar: re.compile(r'^' + str(g_v), re.IGNORECASE)})

        if len(geoq_l) > 1:
            geo_q = {"$and": geoq_l}
        else:
            geo_q = geoq_l[0]

    return geo_q, geo_pars


################################################################################

def return_geo_city_query(geo_root, geo_pars):
    geoq_l = []

    for g_k, g_v in geo_pars.items():

        if len(geo_root) > 0:
            geopar = ".".join([geo_root, "properties", g_k])
        else:
            geopar = ".".join(["properties", g_k])

        geoq_l.append({geopar: re.compile(r'^' + str(g_v), re.IGNORECASE)})

    if len(geoq_l) > 1:
        return {"$and": geoq_l}
    else:
        return geoq_l[0]


################################################################################

def return_geo_longlat_query(geo_root, geo_pars):
    if len(geo_root) > 0:
        geojsonpar = ".".join((geo_root, "geometry"))
    else:
        geojsonpar = "geo_location.geometry"

    geo_q = {
        geojsonpar: {
            '$near': SON(
                [
                    (
                        '$geometry', SON(
                            [
                                ('type', 'Point'),
                                ('coordinates', [
                                    geo_pars["geo_longitude"],
                                    geo_pars["geo_latitude"]
                                ])
                            ]
                        )
                    ),
                    ('$maxDistance', geo_pars["geo_distance"])
                ]
            )
        }
    }

    return geo_q

################################################################################



import pymongo
from os import environ

from cgi_parsing import *
from datatable_utils import get_nested_value
from bycon_helpers import paginate_list
from response_remapping import de_vrsify_variant


################################################################################

def export_variants_download(ds_id, byc):

    data_client = pymongo.MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
    v_coll = data_client[ ds_id ][ "variants" ]
    ds_results = byc["dataset_results"][ds_id]
 
    v__ids = byc["dataset_results"][ds_id]["variants._id"].get("target_values", [])
    if test_truthy( byc["form_data"].get("paginate_results", True) ):
        v__ids = paginate_list(v__ids, byc)

    open_json_streaming(byc, "variants.json")

    for v_id in v__ids[:-1]:
        v = v_coll.find_one( { "_id": v_id }, { "_id": 0 } )
        print(decamelize_words(json.dumps(camelize(v), indent=None, sort_keys=False, default=str, separators=(',', ':')), end = ','))
    v = v_coll.find_one( { "_id": v__ids[-1]}, { "_id": 0 }  )
    print(decamelize_words(json.dumps(camelize(v), indent=None, sort_keys=False, default=str, separators=(',', ':')), end = ''))

    close_json_streaming()


################################################################################

def stream_pgx_meta_header(ds_id, ds_results, byc):

    if not "pgxseg" in byc["output"] and not "pgxmatrix" in byc["output"]:
        return

    s_r_rs = byc["service_response"]["response"]["result_sets"][0]
    b_p = byc["pagination"]

    mongo_client = pymongo.MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
    bs_coll = mongo_client[ ds_id ][ "biosamples" ]

    open_text_streaming(byc["env"])

    for d in ["id", "assemblyId"]:
        print("#meta=>{}={}".format(d, byc["dataset_definitions"][ds_id][d]))
    for k, n in s_r_rs["info"]["counts"].items():
        print("#meta=>{}={}".format(k, n))
    print("#meta=>pagination.skip={};pagination.limit={};pagination.range={},{}".format(b_p["skip"], b_p["skip"], b_p["range"][0] + 1, b_p["range"][1]))
            
    print_filters_meta_line(byc)

    for bs_id in ds_results["biosamples.id"][ "target_values" ]:
        bs = bs_coll.find_one( { "id": bs_id } )
        if not bs:
            continue
        h_line = pgxseg_biosample_meta_line(byc, bs, "histological_diagnosis_id")
        print(h_line)

    return


################################################################################

def pgxseg_biosample_meta_line(byc, biosample, group_id_key="histological_diagnosis_id"):

    dt_m = byc["datatable_mappings"]
    io_params = dt_m["entities"][ "biosample" ]["parameters"]

    g_id_k = group_id_key
    g_lab_k = re.sub("_id", "_label", g_id_k)

    line = [ "#sample=>id={}".format(biosample.get("id", "¡¡¡NONE!!!")) ]

    for par, par_defs in io_params.items():

        in_pgxseg = par_defs.get("compact", False)
        if in_pgxseg is False:
            continue

        parameter_type = par_defs.get("type", "string")
        pres = par_defs.get("prefix_split", {})

        if len(pres.keys()) < 1:
            db_key = par_defs.get("db_key", "___undefined___")
            v = get_nested_value(biosample, db_key)
            h_v = ""
            if isinstance(v, list):
                h_v = "::".join(map(str, (v)))
            else:
                h_v = str(v)

            if len(h_v) > 0:
                if g_id_k == par:
                    line.append("group_id={}".format(h_v))
                if g_lab_k == par:
                    line.append("group_label={}".format(h_v))
                line.append("{}={}".format(par, h_v))
        else:

            par_vals = biosample.get(par, [])
            if not isinstance(par_vals, list):
                continue
            for pre, pre_defs in pres.items():
                in_pgxseg = pre_defs.get("compact", False)
                if in_pgxseg is False:
                    continue
                for v in par_vals:
                    if v.get("id", "___none___").startswith(pre):
                        line.append("{}_id___{}={}".format(par, pre, v.get("id")))
                        l = v.get("label", "")
                        if len(l) > 0:
                            line.append("{}_label___{}={}".format(par, pre, v.get("id")))
                        continue

    h_line = ";".join(line)

    return h_line

################################################################################    

def __pgxmatrix_interval_header(info_columns, byc):

    int_line = info_columns.copy()

    for iv in byc["genomic_intervals"]:
        int_line.append("{}:{}-{}:DUP".format(iv["reference_name"], iv["start"], iv["end"]))
    for iv in byc["genomic_intervals"]:
        int_line.append("{}:{}-{}:DEL".format(iv["reference_name"], iv["start"], iv["end"]))

    return int_line


################################################################################

def print_filters_meta_line(byc):

    if "filters" in byc["service_response"]["meta"]["received_request_summary"]:
        f_vs = []
        for f in byc["service_response"]["meta"]["received_request_summary"]["filters"]:
            f_vs.append(f["id"])
        print("#meta=>filters="+','.join(f_vs))

    return

################################################################################

def export_pgxseg_download(ds_id, byc):

    data_client = pymongo.MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
    v_coll = data_client[ ds_id ][ "variants" ]
    ds_results = byc["dataset_results"][ds_id]
    v__ids = byc["dataset_results"][ds_id]["variants._id"].get("target_values", [])
    if test_truthy( byc["form_data"].get("paginate_results", True) ):
        v__ids = paginate_list(v__ids, byc)

    stream_pgx_meta_header(ds_id, ds_results, byc)
    print_pgxseg_header_line()

    for v_id in v__ids:
        v = v_coll.find_one( { "_id": v_id} )
        print_variant_pgxseg(v, byc)

    close_text_streaming()

################################################################################

def print_variant_pgxseg(v, byc):

    print( pgxseg_variant_line(v, byc) )

################################################################################

def print_pgxseg_header_line():

    print( pgxseg_header_line() )

################################################################################

def pgxseg_header_line():

    return "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}".format("biosample_id", "reference_name", "start", "end", "log2", "variant_type", "reference_bases", "alternate_bases" )

################################################################################

def pgxseg_variant_line(v, byc):

    v = de_vrsify_variant(v, byc)
    if v is False:
        return v

    v["log2"] = v["info"].get("cnv_value", ".")

    return "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}".format(v["biosample_id"], v["reference_name"], v["start"], v["end"], v["log2"], v["variant_type"], v["reference_bases"], v["alternate_bases"])

################################################################################

def export_callsets_matrix(ds_id, byc):

    m_format = "coverage"
    if "val" in byc["output"]:
        m_format = "values"

    ds_results = byc["dataset_results"][ds_id]
    p_r = byc["pagination"]

    if not "callsets._id" in ds_results:
        return

    cs_r = ds_results["callsets._id"]

    mongo_client = pymongo.MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
    bs_coll = mongo_client[ ds_id ][ "biosamples" ]
    cs_coll = mongo_client[ ds_id ][ "callsets" ]

    open_text_streaming(byc["env"], "interval_callset_matrix.pgxmatrix")

    for d in ["id", "assemblyId"]:
        print("#meta=>{}={}".format(d, byc["dataset_definitions"][ds_id][d]))
    print_filters_meta_line(byc)
    print("#meta=>data_format=interval_"+m_format)

    info_columns = [ "analysis_id", "biosample_id", "group_id" ]
    h_line = __pgxmatrix_interval_header(info_columns, byc)
    print("#meta=>genome_binning={};interval_number={}".format(byc["genome_binning"], len(byc["genomic_intervals"])) )
    print("#meta=>no_info_columns={};no_interval_columns={}".format(len(info_columns), len(h_line) - len(info_columns)))

    q_vals = cs_r["target_values"]
    r_no = len(q_vals)
    if r_no > p_r["limit"]:
        if test_truthy( byc["form_data"].get("paginate_results", True) ):
            q_vals = paginate_list(q_vals, byc)
        print('#meta=>"WARNING: Only analyses {} - {} (from {}) will be included pagination skip and limit"'.format((p_r["range"][0] + 1), p_r["range"][-1], cs_r["target_count"]))

    bios_ids = set()
    cs_ids = {}
    cs_cursor = cs_coll.find({"_id": {"$in": q_vals } } )
    for cs in cs_cursor:
        bios = bs_coll.find_one( { "id": cs["biosample_id"] } )
        bios_ids.add(bios["id"])
        s_line = "#sample=>biosample_id={};analysis_id={}".format(bios["id"], cs["id"])
        h_d = bios["histological_diagnosis"]
        cs_ids.update({cs["id"]: h_d.get("id", "NA")})
        s_line = '{};group_id={};group_label={};NCIT::id={};NCIT::label={}'.format(s_line, h_d.get("id", "NA"), h_d.get("label", "NA"), h_d.get("id", "NA"), h_d.get("label", "NA"))
        print(s_line)

    print("#meta=>biosampleCount={};analysisCount={}".format(len(bios_ids), cs_r["target_count"]))
    print("\t".join(h_line))

    for cs_id, group_id in cs_ids.items():
        cs = cs_coll.find_one({"id":cs_id})
        if "values" in m_format:
            print("\t".join(
                [
                    cs_id,
                    cs.get("biosample_id", "NA"),
                    group_id,
                    *map(str, cs["cnv_statusmaps"]["max"]),
                    *map(str, cs["cnv_statusmaps"]["min"])
                ]
            ))
        else:
            print("\t".join(
                [
                    cs_id,
                    cs.get("biosample_id", "NA"),
                    group_id,
                    *map(str, cs["cnv_statusmaps"]["dup"]),
                    *map(str, cs["cnv_statusmaps"]["del"])
                ]
            ))

    close_text_streaming()

################################################################################

def export_pgxseg_frequencies(byc, results):

    if not "pgxseg" in byc["output"] and not "pgxfreq" in byc["output"]:
        return

    open_text_streaming(byc["env"], "interval_frequencies.pgxfreq")

    print("#meta=>genome_binning={};interval_number={}".format(byc["genome_binning"], len(byc["genomic_intervals"])) )
    h_ks = ["reference_name", "start", "end", "gain_frequency", "loss_frequency", "no"]

    # should get error checking if made callable

    for f_set in results:
        m_line = []
        for k in ["group_id", "label", "dataset_id", "sample_count"]:
            m_line.append(k+"="+str(f_set[k]))
        print("#group=>"+';'.join(m_line))

    print("group_id\t"+"\t".join(h_ks))

    for f_set in results:
        for intv in f_set["interval_frequencies"]:
            v_line = [ ]
            v_line.append(f_set[ "group_id" ])
            for k in h_ks:
                v_line.append(str(intv[k]))
            print("\t".join(v_line))

    close_text_streaming()

################################################################################

def export_pgxmatrix_frequencies(byc, results):

    open_text_streaming(byc["env"], "interval_frequencies.pgxmatrix")

    print("#meta=>genome_binning={};interval_number={}".format(byc["genome_binning"], len(byc["genomic_intervals"])) )

    # should get error checking if made callable
    for f_set in results:
        m_line = []
        for k in ["group_id", "label", "dataset_id", "sample_count"]:
            m_line.append(k+"="+str(f_set[k]))
        print("#group=>"+';'.join(m_line))
    # header

    h_line = [ "group_id" ]
    h_line = __pgxmatrix_interval_header(h_line, byc)
    print("\t".join(h_line))

    for f_set in results:
        f_line = [ f_set[ "group_id" ] ]
        for intv in f_set["interval_frequencies"]:
            f_line.append( str(intv["gain_frequency"]) )
        for intv in f_set["interval_frequencies"]:
            f_line.append( str(intv["loss_frequency"]) )

        print("\t".join(f_line))

    close_text_streaming()




import math, re

from cgi_parsing import *
from file_utils import read_www_tsv_to_dictlist

################################################################################

def read_geomarker_table_web(byc):

    geolocs = []

    f_a = byc["form_data"].get("file", "")

    if not "http" in f_a:
        return geolocs

    lf, fieldnames = read_www_tsv_to_dictlist(f_a)

    markers = {}

    for line in lf:
        
        group_lon = line.get("group_lon", "")       # could use 0 here for Null Island...
        group_lat = line.get("group_lat", "")       # could use 0 here for Null Island...
        group_label = line.get("group_label", "")
        item_size = line.get("item_size", "")
        item_label = line.get("item_label", "")
        item_link = line.get("item_link", "")
        marker_type = line.get("marker_type", "circle")
        marker_icon = line.get("marker_icon", "")

        if not re.match(r'^\-?\d+?(?:\.\d+?)?$', str(group_lat) ):
            continue
        if not re.match(r'^\-?\d+?(?:\.\d+?)?$', str(group_lon) ):
            continue
        if not re.match(r'^\d+?(?:\.\d+?)?$', str(item_size) ):
            item_size = 1

        m_k = "{}::LatLon::{}::{}".format(line["group_label"], line["group_lat"], line["group_lon"])

        # TODO: load schema for this
        if not m_k in markers.keys():
            markers[m_k] = {
                "geo_location": {
                    "type": "Feature",
                    "geometry": {
                        "type": "Point",
                        "coordinates": [ float(group_lon), float(group_lat) ]
                    },
                    "properties": {
                        "city": None,
                        "country": None,
                        "label": group_label,
                        "marker_type": marker_type,
                        "marker_icon": marker_icon,
                        "marker_count": 0,
                        "items": []
                    }
                }
            }

        g_l_p = markers[m_k]["geo_location"]["properties"]
        g_l_p["marker_count"] += float(item_size)

        if len(item_label) > 0:
            if "http" in item_link:
                item_label = "<a href='{}' target='_blank'>{}</a>".format(item_link, item_label)
            g_l_p["items"].append(item_label)

    for m_k, m_v in markers.items():
        geolocs.append(m_v)

    return geolocs

################################################################################

def print_map_from_geolocations(byc, geolocs):

    if not "map" in byc["output"]:
        return

    m_p = byc["geoloc_definitions"].get("map_params", {})
    p_p = update_plot_params_from_form(byc)
    m_max_count = marker_max_from_geo_locations(geolocs)
    
    # print(m_max_count)
    leaf_markers = []
    for geoloc in geolocs:
        leaf_markers.append( map_marker_from_geo_location(byc, geoloc, p_p, m_max_count) )
    markersJS = create_marker_layer(leaf_markers)

    geoMap = """
<!-- map needs to exist before we load leaflet -->
{}
<div id="{}" style="width: {}px; height: {}px;"></div>

<!-- Make sure you put this AFTER Leaflet's CSS -->
<script src="https://unpkg.com/leaflet@1.8.0/dist/leaflet.js"
      integrity="sha512-BB3hKbKWOc9Ez/TAwyWxNXeoV9c1v6FIeYiBieIWkpLjauysF18NzgR1MBNBXf8/KABdlkX68nAhlwcDFLGPCQ=="
      crossorigin=""></script>
<script>
  var circleOptions = {{
    color: '{}',
    stroke: true,
    weight: {},
    fillColor: '{}',
    fillOpacity: {},
    radius: 1000,
    count: 1
  }};

  // Create the map.

  var map = L.map('{}', {{ renderer: L.svg() }} ).setView([{}, {}], {});

  L.tileLayer('{}', {{
      minZoom: {},
      maxZoom: {},
      attribution: '{}'
  }}).addTo(map);

{}

</script>
    """.format(
        m_p.get("head"),
        m_p.get("plotid"),
        p_p.get("map_w_px"),
        p_p.get("map_h_px"),
        p_p["bubble_stroke_color"],
        p_p["bubble_stroke_weight"],
        p_p["bubble_fill_color"],
        p_p["bubble_opacity"],
        m_p.get("plotid"),
        m_p.get("init_latitude"),
        m_p.get("init_longitude"),
        m_p.get("zoom"),
        m_p.get("tiles_source"),
        p_p.get("zoom_min"),
        p_p.get("zoom_max"),
        m_p.get("attribution"),
        markersJS
    )

    if test_truthy(byc["form_data"].get("help", False)):
        t = """
<h4>Map Configuration</h4>
<p>The following parameters may be modified by providing alternative values in
the URL, e.g. "&map_w_px=1024".</p>
<table>
"""
        t += "<tr><th>Map Parameter</th><th>Value</th></tr>\n"
        for p_p_k, p_p_v in p_p.items():
            if not '<' in str(p_p_v):
                t += "<tr><td>{}</td><td>{}</td></tr>\n".format(p_p_k, p_p_v)
        t += "\n</table>"
        geoMap += t

    print("""
<html>
{}
</html>""".format(geoMap))

    exit()

################################################################################

def create_marker_layer(leaf_markers):

    markersJS = ""

    if len(leaf_markers) > 0:
        markersJS = """
  var markers = [
{}
  ];
  var markersGroup = L.featureGroup(markers);
  map.addLayer(markersGroup);
  map.fitBounds(markersGroup.getBounds().pad(0.05));
""".format(",\n".join(leaf_markers))

    return markersJS

################################################################################

def update_plot_params_from_form(byc):

    p_p = byc["geoloc_definitions"].get("plot_params", {})

    p_p.update({"file": byc["form_data"].get("file", "")})

    for p_p_k, p_p_v in p_p.items():

        if p_p_k in byc["form_data"]:
            p_p.update({p_p_k: byc["form_data"].get(p_p_k, p_p_v)})

    return p_p

################################################################################

def marker_max_from_geo_locations(geolocs):

    m_max_count = 1
    for g_l in geolocs:
        c = float( g_l["geo_location"]["properties"].get("marker_count", 1) )
        if c > m_max_count:
            m_max_count = c

    return m_max_count

################################################################################

def map_marker_from_geo_location(byc, geoloc, p_p, m_max_count):

    p = geoloc["geo_location"]["properties"]
    g = geoloc["geo_location"]["geometry"]

    marker = p_p.get("marker_type", "circle")
    # if m_max_count == 1:
    #     marker = "marker"



    m_max_r = p_p.get("marker_max_r", 1000)
    m_f = int(int(m_max_r) / math.sqrt(4 * m_max_count / math.pi))

    label = p.get("label", None)
    if label is None:
        label = p.get("city", "NA")
        country = p.get("country", None)
        if country is not None:
            label = "{}, {}".format(label, country)

    items = p.get("items", [])
    items = [x for x in items if x is not None]
    # print(items)
    if len(items) > 0:
        label += "<hr/>{}".format("<br/>".join(items))
    else:
        label += "<hr/>latitude: {}, longitude: {}".format(g["coordinates"][1], g["coordinates"][0])

    count = float(p.get("marker_count", 1))
    size = count * m_f * float(p_p.get("marker_scale", 2))

    marker_icon = p.get("marker_icon", "")

    if ".png" in marker_icon or ".jpg" in marker_icon:
        marker = "marker"

    if "circle" in marker:
        map_marker = """
L.{}([{}, {}], {{
    ...circleOptions,
    ...{{radius: {}, count: {}}}
}}).bindPopup("{}", {{maxHeight: 200}})
    """.format(
        marker,
        g["coordinates"][1],
        g["coordinates"][0],
        size,
        count,
        label
    )

    else:
        map_marker = """
L.{}([{}, {}], {{
    ...{{count: {}}}
}}).bindPopup("{}", {{maxHeight: 200}})
    """.format(
        marker,
        g["coordinates"][1],
        g["coordinates"][0],
        count,
        label
    )

    return map_marker



import scipy.cluster

################################################################################

def cluster_frequencies(plv, byc):

    m = plv.get("plot_cluster_metric", "complete")

    matrix = matrix_from_interval_frequencies(plv, byc)
    linkage = scipy.cluster.hierarchy.linkage(matrix, method=m)
    dendrogram = scipy.cluster.hierarchy.dendrogram(linkage, no_plot=True, orientation="right")
    
    return dendrogram

################################################################################

def matrix_from_interval_frequencies(plv, byc):

    matrix = []

    for f_set in plv["results"]:

        i_f = f_set.get("interval_frequencies", [])
        if_line = []

        for i_f in f_set.get("interval_frequencies", []):
            if_line.append( i_f.get("gain_frequency", 0) )
        for i_f in f_set.get("interval_frequencies", []):
            if_line.append( i_f.get("loss_frequency", 0) )

        matrix.append(if_line)

    return matrix

################################################################################

def cluster_samples(plv, byc):

    m = plv.get("plot_cluster_metric", "complete")

    matrix = []

    for s in plv["results"]:
        s_line = []

        if "intcoverage" in plv.get("plot_samples_cluster_type", ""):

            c_m = s.get("cnv_statusmaps", {})
            dup_l = c_m.get("dup", [])
            del_l = c_m.get("del", [])
            for i_dup in dup_l:
                s_line.append(i_dup)
            for i_del in del_l:
                s_line.append(i_del)
        else:
            c_s = s.get("cnv_chro_stats", {})
            for c_a, c_s_v in c_s.items():
                s_line.append(c_s_v.get("dupfraction", 0))
            for c_a, c_s_v in c_s.items():
                s_line.append(c_s_v.get("delfraction", 0))

        matrix.append(s_line)

    linkage = scipy.cluster.hierarchy.linkage(matrix, method=m)
    reorder = scipy.cluster.hierarchy.leaves_list(linkage)
    dendrogram = scipy.cluster.hierarchy.dendrogram(linkage, no_plot=True, orientation="right")

    return dendrogram



import re

from cgi_parsing import boolean_to_mongo_logic

################################################################################

def parse_filters(byc):

    get_global_filter_flags(byc)
    byc.update( { "filters": check_filter_values(byc) } )


################################################################################

def get_global_filter_flags(byc):

    ff = {
        "logic": byc[ "config" ][ "filter_flags" ][ "logic" ],
        "precision": byc[ "config" ][ "filter_flags" ][ "precision" ],
        "descendants": byc[ "config" ][ "filter_flags" ][ "include_descendant_terms" ]
    }

    if "form_data" in byc:
        if "filter_logic" in byc[ "form_data" ]:
            ff["logic"] = boolean_to_mongo_logic( byc["form_data"]['filter_logic'] )
        if "filter_precision" in byc[ "form_data" ]:
            ff["precision"] = byc["form_data"]['filter_precision']
        if "include_descendant_terms" in byc[ "form_data" ]:
            i_d_t = str(byc[ "form_data" ].get("include_descendant_terms", 1)).lower()
            if i_d_t in ["0", "-1", "no", "false"]:
                ff["descendants"] = False

    byc.update( { "filter_flags": ff } )

################################################################################

def check_filter_values(byc):

    """
    The function checks the filter values for a match to any of the filter
    definitions. The optional `!` flag (no match) is not considered during
    evaluation ("deflagged").
    This filter check is complementary to the evaluation during the filter query
    generation and provides a warning if the filter pattern doesn't exist.
    """

    checked = [ ]
    filters = byc["form_data"].get("filters")
    if not filters:
        return checked

    f_defs = byc["filter_definitions"]

    for f in filters:
        if not isinstance(f, dict):
            f = {"id":f}
        if not "id" in f:
            continue

        deflagged = re.sub(r'^!', '', f["id"])
        matched = False
        for f_t, f_d in f_defs.items():
            if re.compile( f_d["pattern"] ).match( deflagged ):
                # print(f'{f_d["pattern"]} => ? {deflagged}')
                matched = True
                continue

        if matched is False:
            warning = "The filter `{}` does not match any defined filter pattern.".format(f["id"])
            response_add_filter_warnings(byc, warning)

        if f not in checked:
            checked.append( f )

    return checked


################################################################################

def response_add_filter_warnings(byc, message=False):

    if message is False:
        return

    if len(str(message)) < 1:
        return

    if not "service_response" in byc:
        return

    if not "info" in byc["service_response"] or byc["service_response"]["info"] is None:
        byc["service_response"].update({"info": {"warnings":[]}})
    if not "warnings" in byc["service_response"]["info"] or byc["service_response"]["info"]["warnings"] is None:
        byc["service_response"]["info"].update({"warnings": []})

    byc["service_response"]["info"]["warnings"].append(message)



import cgi, json, re, sys
from urllib.parse import urlparse, parse_qs, unquote
from os import environ
from humps import camelize, decamelize


################################################################################

def set_debug_state(debug: int = 0) -> bool:
    """
    Function to provide a text response header for debugging purposes, i.e. to 
    print out the error or test parameters to a browser session.
    The common way would be to add either a `/debug=1/` part to a REST path or
    to provide a `...&debug=1` query parameter.
    """

    if test_truthy(debug):
        print('Content-Type: text')
        print()
        return True

    r_uri = environ.get('REQUEST_URI', "___none___")
    if re.match(r'^.*?[?&/]debug=(\w+?)\b.*?$', r_uri):
        d = re.match(r'^.*?[?&/]debug=(\w+?)\b.*?$', r_uri).group(1)
        if test_truthy(d):
            print('Content-Type: text')
            print()
            return True

    return False


################################################################################

def boolean_to_mongo_logic(logic: str = "AND") -> str:
    if "OR" in logic.upper():
        return '$or'
    return '$and'


################################################################################

def select_this_server(byc: dict) -> str:
    s_uri = str(environ.get('SCRIPT_URI'))

    # TODO
    if "progenetix.test" in s_uri:
        if "https:" in s_uri:
            return "https://progenetix.test"
        else:
            return "http://progenetix.test"
    elif "cancercelllines.test" in s_uri:
        if "https:" in s_uri:
            return "https://cancercelllines.test"
        else:
            return "http://cancercelllines.test"

            # TODO: ERROR hack for https/http mix, CORS...
    # ... since cloudflare provides https mapping
    # if "https:" in s_uri:
    if "http:" in s_uri:
        return "https://" + str(environ.get('HTTP_HOST'))
    else:
        return "http://" + str(environ.get('HTTP_HOST'))


################################################################################

def cgi_parse_query(byc):
    r_m = environ.get('REQUEST_METHOD', '')
    if "POST" in r_m:
        cgi_parse_POST(byc)
    else:
        cgi_parse_GET(byc)


################################################################################

def cgi_parse_POST(byc):
    content_len = environ.get('CONTENT_LENGTH', '0')
    content_typ = environ.get('CONTENT_TYPE', '')

    b_defs = byc.get("beacon_defaults", {})
    form = {}

    # TODO: catch error & return for non-json posts
    if "json" in content_typ:
        body = sys.stdin.read(int(content_len))
        jbod = json.loads(body)
        if "debug" in jbod:
            if jbod["debug"] > 0:
                byc.update({"debug_mode": set_debug_state(1)})

        # TODO: this hacks the v2 structure; ideally should use requestParameters schemas
        if "query" in jbod:
            for p, v in jbod["query"].items():
                if p == "requestParameters":
                    for rp, rv in v.items():
                        rp_d = decamelize(rp)
                        if "datasets" in rp:
                            if "datasetIds" in rv:
                                form.update({"dataset_ids": rv["datasetIds"]})
                        elif "g_variant" in rp:
                            for vp, vv in v[rp].items():
                                vp_d = decamelize(vp)
                                form.update({vp_d: vv})
                        else:
                            form.update({rp_d: rv})
                else:
                    p_d = decamelize(p)
                    form.update({p_d: v})

        # TODO: define somewhere else with proper defaults
        form.update({
            "requested_granularity": jbod.get("requestedGranularity", b_defs.get("requested_granularity", "record")),
            "include_resultset_responses": jbod.get("includeResultsetResponses",
                                                    b_defs.get("include_resultset_responses", "HIT")),
            "include_handovers": jbod.get("includeHandovers", b_defs.get("include_handovers", False)),
            "filters": jbod.get("filters", [])
        })

        # transferring pagination where existing to standard form values
        pagination = jbod.get("pagination", {})
        for p_k in ["skip", "limit"]:
            if p_k in pagination:
                if re.match(r'^\d+$', str(pagination[p_k])):
                    form.update({p_k: pagination[p_k]})
        byc.update({
            "form_data": form,
            "query_meta": jbod.get("meta", {})
        })


################################################################################

def cgi_parse_GET(byc):
    b_defs = byc.get("beacon_parameters", {})
    p_defs = byc.get("plot_defaults", {})
    v_defs = byc.get("variant_parameters", {})

    f_defs = {**b_defs.get("parameters", {}), **p_defs.get("parameters", {}), **v_defs.get("parameters", {})}

    form = {}

    byc.update({"debug_mode": set_debug_state()})
    get = cgi.FieldStorage()

    for p in get:
        p_d = decamelize(p)
        if p_d in f_defs:
            form.update({p_d: refactor_value_from_defined_type(p, get, f_defs[p_d])})
        # TODO still fallback ..
        else:
            v = get.getvalue(p)
            # making sure double entries are forced to single
            if type(v) is list:
                form.update({p_d: v[0]})
            else:
                form.update({p_d: v})

    # TODO: re-evaluate hack of empty filters which avoids dirty errors downstream
    if not "filters" in form:
        form.update({"filters": []})

    form.update({
        "requested_granularity": get.getvalue("requestedGranularity", b_defs.get("requested_granularity", "record")),
        "include_resultset_responses": get.getvalue("includeResultsetResponses",
                                                    b_defs.get("include_resultset_responses", "HIT")),
        "include_handovers": get.getvalue("includeHandovers", b_defs.get("include_handovers", False))
    })

    if "requested_schema" in form:
        try:
            byc["query_meta"].update({
                "requested_schemas": [{"entity_type": form["requested_schema"]}]
            })
        except:
            pass

    byc.update({"form_data": form})


################################################################################

def rest_path_elements(byc):
    """
    The function deparses a Beacon REST path into its components and assigns
    those to the respective variables. The assumes structure is:

    `__root__/__request-entity__/__entity-id__/__response-entity__/?query...`
        |             |                 |               |
    "beacon"  e.g. "biosamples"  "pgxbs-t4ee3"  e.g. "genomicVariations"
        |             |                 |               |
    required      required          optional        optional
    """

    r_p_r = byc.get("request_path_root", "beacon")

    if not environ.get('REQUEST_URI'):
        return

    url_comps = urlparse(environ.get('REQUEST_URI'))
    url_p = url_comps.path
    p_items = re.split('/', url_p)

    if not r_p_r in p_items:
        return

    for d_k in ["debug=1", "&debug=1", "debug=true"]:
        if d_k in p_items:
            p_items.remove(d_k)

    p_items = list(filter(None, p_items))

    r_i = p_items.index(r_p_r)

    for p_k in ["request_entity_path_id", "request_entity_path_id_value", "response_entity_path_id"]:

        r_i += 1
        if r_i >= len(p_items):
            return
        # print(f'{p_k}: {unquote(p_items[r_i])}')
        byc.update({p_k: unquote(p_items[r_i])})


################################################################################

def rest_path_value(key=""):
    """
    This function splits the path of the REQUEST_URI and returns the path element
    after a provided key. The typical uise case would be to get the entity or
    executing script, or an {id} value from a REST path e.g.

    * `/beacon/biosamples/?` => "beacon" -> "biosamples"
    * `/services/cytomapper/?` => "services" -> "cytomapper"
    * `/services/intervalFrequencies/NCIT:C3072/` => "intervalFrequencies" -> "NCIT:C3072"

    """

    if not environ.get('REQUEST_URI'):
        return

    url_comps = urlparse(environ.get('REQUEST_URI'))
    p_items = re.split('/', url_comps.path)
    p_items = [x for x in p_items if len(x) > 1]
    p_items = [x for x in p_items if not "debug=" in x]

    for i, p in enumerate(p_items, 1):
        if len(p_items) > i:
            if unquote(p) in [key, f'{key}.py', unquote(key)]:
                return unquote(p_items[i])
        elif p == key:
            return

    return


################################################################################

def refactor_value_from_defined_type(parameter, form_data, definition):
    p_d_t = definition.get("type", "string")

    if "array" in p_d_t:
        values = form_return_listvalue(form_data, parameter)

        p_i_t = definition.get("items", "string")

        if "int" in p_i_t:
            return list(map(int, values))
        elif "number" in p_i_t:
            return list(map(float, values))
        else:
            return list(map(str, values))

    else:
        value = form_data.getvalue(parameter)

        if "int" in p_d_t:
            return int(value)
        elif "number" in p_d_t:
            return float(value)
        elif "bool" in p_d_t:
            return test_truthy(value)
        else:
            return str(value)


################################################################################

def form_return_listvalue(form_data, parameter):
    l_v = []
    if len(form_data) > 0:
        if parameter in form_data:
            v = form_data.getlist(parameter)
            if "null" in v:
                v.remove("null")
            if "undefined" in v:
                v.remove("undefined")
            if len(v) > 0:
                l_v = ','.join(v)
                l_v = l_v.split(',')

    return l_v


################################################################################

def test_truthy(this):
    if str(this).lower() in ["1", "true", "y", "yes"]:
        return True

    return False


################################################################################

def get_plot_parameters(plv, byc):
    """
    The new version here assumes that the values in byc["form_data"] already have
    been processed into the correct type.
    """

    p_d_p = byc["plot_defaults"]["parameters"]
    p_d_l = byc["plot_defaults"]["legacy_parameters"]
    form = byc["form_data"]

    for p_k, p_d in p_d_p.items():

        if p_k in form:
            plv.update({p_k: form[p_k]})
        l_p = p_d_l.get(p_k, False)
        if l_p is not False:
            if l_p in byc["form_data"]:
                plv.update({p_k: form[l_p]})

    return plv


################################################################################

def cgi_simplify_response(byc):
    r = byc["service_response"].get("response", "ERROR: No response element in error_response")

    if "result_sets" in r:
        r_s = r["result_sets"][0]
        byc.update({"service_response": r_s.get("results", [])})
    else:
        byc.update({"service_response": r})


################################################################################

def cgi_break_on_errors(byc):
    e_c = byc["error_response"]["error"].get("error_code", 200)
    # if byc.get("debug_mode", False) is True:
    #     prjsonnice(byc["error_response"])

    # TODO: temp hack
    for k in byc["service_response"].keys():
        if isinstance(byc["service_response"][k], dict):
            if "any_of" in byc["service_response"][k]:
                byc["service_response"][k].pop("any_of")
            if "all_of" in byc["service_response"][k]:
                byc["service_response"][k].pop("all_of")

    if int(e_c) > 200:
        cgi_print_response(byc, e_c)


################################################################################

def cgi_debug_message(byc, label, debug_object):
    try:
        if byc["debug_mode"]:
            print("{}:\n\n{}\n\n".format(label, debug_object))
    except:
        pass


################################################################################

def switch_to_wrong_service_response(byc):
    byc.update({
        "service_response": {
            "response": {
                "error": {
                    "error_code": 422,
                    "error_message": "No correct service path provided. Please refer to the documentation at http://info.progenetix.org/tags/Beacon"
                },
            }
        }
    })


################################################################################

def cgi_print_response(byc, status_code):
    r_f = ""
    f_d = {}

    delint_response(byc)

    if "form_data" in byc:
        f_d = byc["form_data"]

    if "responseFormat" in f_d:
        r_f = f_d["responseFormat"]

    # This is a simple "de-jsonify", intended to be used for already
    # pre-formatted list-like items (i.e. lists only containing objects)
    # with simple key-value pairs)
    # TODO: universal text table converter ... partially implemented

    if "text" in byc["output"]:

        cgi_simplify_response(byc)

        if isinstance(byc["service_response"], dict):
            # TODO: Find where this results/response ambiguity comes from
            if "response" in byc["service_response"]:
                resp = byc["service_response"]["response"]
            elif "results" in byc["service_response"]:
                resp = byc["service_response"]["results"]
            else:
                resp = byc["service_response"]
        else:
            resp = byc["service_response"]
        if isinstance(resp, dict):
            resp = json.dumps(camelize(resp), default=str)
        else:
            l_d = []
            for dp in resp:
                v_l = []
                for v in dp.values():
                    # print(v)
                    v_l.append(str(v))
                l_d.append("\t".join(v_l))
            resp = "\n".join(l_d)

        print_text_response(resp, byc["env"], status_code)

    if test_truthy(byc["form_data"].get("only_handovers", False)):
        try:
            if "result_sets" in byc["service_response"]["response"]:
                for rs_i, rs in enumerate(byc["service_response"]["response"]["result_sets"]):
                    byc["service_response"]["response"]["result_sets"][rs_i].update({"results": []})
        except:
            pass

    if "simple" in r_f:
        cgi_simplify_response(byc)

    update_error_code_from_response_summary(byc)
    switch_to_error_response(byc)

    print_json_response(byc["service_response"], byc["env"])


################################################################################

def update_error_code_from_response_summary(byc):
    if not "response_summary" in byc["service_response"]:
        return

    if not "exists" in byc["service_response"]["response_summary"]:
        return


################################################################################

def switch_to_error_response(byc):
    e_c = byc["error_response"]["error"].get("error_code", 200)

    if e_c > 200:
        if "meta" in byc["service_response"]:
            byc["error_response"].update({"meta": byc["service_response"]["meta"]})
        byc["service_response"] = byc["error_response"]


################################################################################

def check_switch_to_boolean_response(byc):
    try:
        if "boolean" in byc["service_response"]["meta"]["received_request_summary"].get("requested_granularity",
                                                                                        "record"):
            byc["service_response"].pop("response", None)
            byc["service_response"]["response_summary"].pop("num_total_results", None)
            byc["service_response"]["meta"].update({"returned_granularity": "boolean"})
    except:
        pass


################################################################################

def check_switch_to_count_response(byc):
    try:
        if "count" in byc["service_response"]["meta"]["received_request_summary"].get("requested_granularity",
                                                                                      "record"):
            byc["service_response"].pop("response", None)
            byc["service_response"]["meta"].update({"returned_granularity": "count"})
    except:
        pass


################################################################################

def delint_response(byc):

    b_s_r = byc["service_response"]

    byc.update({"service_response": response_delete_none_values(b_s_r)})

    try:
        b_h_r = b_s_r.get("beacon_handovers", [])
        if len(b_h_r) < 1:
            byc["service_response"].pop("beacon_handovers", None)
        if not "url" in b_h_r[0]:
            byc["service_response"].pop("beacon_handovers", None)
    except:
        pass

################################################################################

def response_delete_none_values(response):
    """Delete None values recursively from all of the dictionaries"""

    for key, value in list(response.items()):
        if isinstance(value, dict):
            response_delete_none_values(value)
        elif value is None:
            del response[key]
        elif isinstance(value, list):
            for v_i in value:
                if isinstance(v_i, dict):
                    response_delete_none_values(v_i)

    return response

################################################################################
################################################################################
################################################################################

def open_json_streaming(byc, filename="data.json"):
    meta = byc["service_response"].get("meta", {})

    if not "local" in byc["env"]:
        print_json_download_header(filename)

    print('{"meta":', end='')
    print(json.dumps(camelize(meta), indent=None, sort_keys=True, default=str), end=",")
    print('"response":{', end='')
    for r_k, r_v in byc["service_response"].items():
        if "results" in r_k:
            continue
        if "meta" in r_k:
            continue
        print('"' + r_k + '":', end='')
        print(json.dumps(camelize(r_v), indent=None, sort_keys=True, default=str), end=",")
    print('"results":[', end="")


################################################################################

def print_json_download_header(filename):
    print('Content-Type: application/json')
    print(f'Content-Disposition: attachment; filename="{filename}"')
    print('status: 200')
    print()


################################################################################

def close_json_streaming():
    print(']}}')
    exit()


################################################################################

def open_text_streaming(env="server", filename="data.pgxseg"):
    if not "local" in env:
        print('Content-Type: text/plain')
        if not "browser" in filename:
            print('Content-Disposition: attachment; filename="{}"'.format(filename))
        print('status: 200')
        print()


################################################################################

def close_text_streaming():
    print()
    exit()


################################################################################

def prjsoncam(this):
    prjsonnice(camelize(this))


################################################################################

def prjsonnice(this):
    print(decamelize_words(json.dumps(this, indent=4, sort_keys=True, default=str)) + "\n")


################################################################################

def decamelize_words(j_d):
    de_cams = ["gVariants", "sequenceId", "relativeCopyClass", "speciesId", "chromosomeLocation", "genomicLocation"]
    for d in de_cams:
        j_d = re.sub(r"\b{}\b".format(d), decamelize(d), j_d)

    return j_d


################################################################################

def print_json_response(this={}, env="server", status_code=200):
    if not "local" in env:
        print('Content-Type: application/json')
        print('status:' + str(status_code))
        print()

    prjsoncam(this)
    print()
    exit()


################################################################################

def print_svg_response(this, env="server", status_code=200):
    if not this:
        return

    if "server" in env:
        print('Content-Type: image/svg+xml')
        print('status:' + str(status_code))
        print()

    elif "file" in env:
        # this opion can be used to reroute the response to a file
        return this

    print(this)
    print()
    exit()


################################################################################

def print_text_response(this="", env="server", status_code=200):
    if "server" in env:
        print('Content-Type: text/plain')
        print('status:' + str(status_code))
        print()

    elif "file" in env:
        # this opion can be used to reroute the response to a file
        return this

    print(this)
    print()
    exit()


################################################################################

def print_uri_rewrite_response(uri_base="", uri_stuff=""):
    print("Status: 302")
    print("Location: {}{}".format(uri_base, uri_stuff))
    print()
    exit()



import base36
import time
import re
from isodate import parse_duration

################################################################################

def generate_id(prefix):
    time.sleep(.001)
    return '{}-{}'.format(prefix, base36.dumps(int(time.time() * 1000)))  ## for time in ms


################################################################################

def days_from_iso8601duration(iso8601duration):
    """A simple function to convert ISO8601 duration strings to days. This is
    potentially lossy since it does not include time parsing."""

    # TODO: check format
    is_isodate_duration = re.match(r'^P\d+?[YMD](\d+?[M])?(\d+?[D])?', iso8601duration)
    if not is_isodate_duration:
        return False

    duration = parse_duration(iso8601duration)
    days = 0
    try:
        days += int(duration.years) * 365.2425
    except AttributeError:
        pass
    try:
        days += int(duration.months) * 30.4167
    except AttributeError:
        pass
    try:
        days += int(duration.days)
    except AttributeError:
        pass

    return days


################################################################################

def hex_2_rgb( hexcolor ):

    rgb = [127, 127, 127]
    h = hexcolor.lstrip('#')
    rgb = tuple(int(h[i:i+2], 16) for i in (0, 2, 4))

    return rgb

################################################################################

def set_pagination_range(d_count, byc):
    r_range = [
        byc["pagination"]["skip"] * byc["pagination"]["limit"],
        byc["pagination"]["skip"] * byc["pagination"]["limit"] + byc["pagination"]["limit"],
    ]

    if byc["pagination"]["skip"] == 0 and byc["pagination"]["limit"] == 0:
        byc["pagination"].update({"range": [0, d_count]})
        return

    r_l_i = d_count - 1

    if r_range[0] > r_l_i:
        r_range[0] = r_l_i
    if r_range[-1] > d_count:
        r_range[-1] = d_count

    byc["pagination"].update({"range": r_range})


################################################################################

def paginate_list(this, byc):
    if byc["pagination"]["limit"] < 1:
        return this

    r = byc["pagination"]["range"]

    t_no = len(this)
    r_min = r[0] + 1
    r_max = r[-1]

    if r_min > t_no:
        return []
    if r_max > t_no:
        return this[r[0]:r_max]

    return this[r[0]:r[-1]]






import re, json, yaml

from json_ref_dict import RefDict, materialize
from humps import decamelize
from os import path, scandir, pardir
from pathlib import Path

from cgi_parsing import prjsonnice

# local
lib_path = path.dirname( path.abspath(__file__) )
pkg_path = path.join( lib_path, pardir )

################################################################################

def read_schema_file(schema_name, item, byc, ext="json"):

    b_p_m = byc["beacon_mappings"]["default_schema_from_model"]

    schema_name = b_p_m.get(schema_name, schema_name)
    
    s_f_p = get_schema_file_path(schema_name, ext)

    if byc["debug_mode"] is True:
        print(schema_name, s_f_p)

    if s_f_p is not False:
        if len(item) > 1:
            s_f_p = s_f_p+"#/"+item
        root_def = RefDict(s_f_p)
        
        exclude_keys = [ "examples" ] #"format",
        s = materialize(root_def, exclude_keys=exclude_keys)
        assert isinstance(s, dict)
        return s

    return False

################################################################################

def get_schema_file_path(schema_name, ext="json"):

    e_l = f'.{ext}'
    d_n = f'defaultSchema.{ext}'
    s_n = f'{schema_name}.{ext}'

    p = Path(path.join( pkg_path, "schemas" ))
    s_p_s = [ f for f in p.rglob("*") if f.is_file() ]
    s_p_s = [ f for f in s_p_s if f.suffix == e_l ]

    for f_p in s_p_s:
        if f_p.parent.name == schema_name:
            if f_p.name == d_n:
                return f'{f_p.resolve()}'
        elif f_p.name == s_n:
            return f'{f_p.resolve()}'

    return False


################################################################################

def instantiate_schema(schema):
    if 'type' in schema:
        if schema['type'] == 'object' and 'properties' in schema:
            empty_dict = {}
            for prop, prop_schema in schema['properties'].items():
                empty_dict[prop] = instantiate_schema(prop_schema)
            return empty_dict
        elif schema['type'] == 'array' and 'items' in schema:
            return [instantiate_schema(schema['items'])]
        elif "const" in schema:
            return schema.get("const", "")
        elif "default" in schema:
            return schema["default"]

    return None
  

################################################################################

def create_empty_instance(schema):
    s_i = instantiate_schema(schema)
    s_i = decamelize(s_i)
    return s_i

################################################################################

def object_instance_from_schema_name(byc, schema_name, root_key, ext="json"):
    s_f = read_schema_file(schema_name, root_key, byc, ext)
    s_i = create_empty_instance( s_f )

    return s_i

################################################################################



import inspect
import os
from pathlib import Path

from args_parsing import *
from bycon_plot import ByconPlot
from bycon_helpers import paginate_list, set_pagination_range
from cgi_parsing import prjsonnice
from data_retrieval import *
from dataset_parsing import select_dataset_ids
from datatable_utils import export_datatable_download
from export_file_generation import *
from file_utils import ByconBundler, callset_guess_probefile_path
from filter_parsing import parse_filters
from genome_utils import translate_reference_ids
from handover_generation import dataset_response_add_handovers, query_results_save_handovers, \
    dataset_results_save_handovers
from interval_utils import generate_genomic_mappings
from query_execution import execute_bycon_queries, mongo_result_list
from query_generation import generate_dataset_queries
from read_specs import load_yaml_empty_fallback, read_bycon_definition_files, read_local_prefs
from response_remapping import *
from variant_parsing import parse_variants
from schema_parsing import *

################################################################################

def initialize_bycon(config):

    b_r_p = config.get("byc_root_pars", {})

    bycon_lib_path = os.path.dirname(os.path.abspath(__file__))
    pkg_path = os.path.join(bycon_lib_path, os.pardir)

    byc = {
        "pkg_path": pkg_path,
        "bycon_lib_path": bycon_lib_path
    }

    for k, v in b_r_p.items():
        byc.update({k: v})

    config.pop("byc_root_pars", None)
    byc.update({"config": config})

    if not environ.get('HTTP_HOST'):
        byc.update({"env": "local"})

    return byc


################################################################################

def beacon_data_pipeline(byc, entry_type):
    initialize_bycon_service(byc, entry_type)

    run_beacon_init_stack(byc)
    return_filtering_terms_response(byc)
    run_result_sets_beacon(byc)
    update_meta_queries(byc)
    query_results_save_handovers(byc)
    check_callset_plot_delivery(byc)
    check_computed_histoplot_delivery(byc)
    check_computed_interval_frequency_delivery(byc)
    check_switch_to_count_response(byc)
    check_switch_to_boolean_response(byc)
    cgi_print_response(byc, 200)


################################################################################

def run_beacon_init_stack(byc):
    select_dataset_ids(byc)
    if len(byc["dataset_ids"]) < 1:
        print_text_response("No existing dataset_id - please check dataset_definitions")

    create_empty_beacon_response(byc)
    parse_filters(byc)
    parse_variants(byc)
    response_add_received_request_summary_parameters(byc)
    generate_genomic_mappings(byc)
    response_collect_errors(byc)
    cgi_break_on_errors(byc)


################################################################################

def initialize_bycon_service(byc, service=False):
    """For consistency, the name of the local configuration file should usually
    correspond to the calling main function. However, an overwrite can be
    provided."""

    form = byc["form_data"]
    scope = "beacon"

    if not service:
        service = byc.get("request_entity_path_id", False)
    frm = inspect.stack()[1]
    if not service:
        service = frm.function

    # TODO - streamline, also for services etc.
    s_a_s = byc["beacon_mappings"].get("service_aliases", {})

    if service in s_a_s:
        service = s_a_s[service]

    mod = inspect.getmodule(frm[0])

    if mod is not None:
        """
        Here we allow the addition of additional configuration files, necessary
        for options beyond basic library use. Files are read in as
        1. from a `local` directory inside the script directory
            * this is the location of configuration file w/ content differing on
              the beaconServer instance
            * these files are inserted during installation (see the documentation)
        2. from a `config` directory inside the script directory
            * script specific configurations are stored there under the name of
              script, e.g. `frequencymaps_creator.yaml` - typically only
              used in services and utility scripts, not beaconServer
            * this is usually used to provide script-specific parameters
              (`service_defaults`...)
        """
        sub_path = path.dirname(path.abspath(mod.__file__))
        if "services" in sub_path or "byconaut" in sub_path:
            scope = "services"
        byc.update({
            "request_path_root": scope,
            "request_entity_path_id": service
        })

        loc_dir = path.join(sub_path, "local")
        read_bycon_definition_files(loc_dir, byc)
        defaults = byc["beacon_defaults"].get("defaults", {})
        for d_k, d_v in defaults.items():
            byc.update({d_k: d_v})
    
        read_local_prefs(service, sub_path, byc)

    get_bycon_args(byc)
    args_update_form(byc)

    if "defaults" in byc.get("service_config", {}):
        for d_k, d_v in byc["service_config"]["defaults"].items():
            byc.update({d_k: d_v})

    d_k = f'{scope}_defaults'
    defs = byc.get(d_k, {})
    b_e_d = defs.get("entity_defaults", {})

    snaked_service = decamelize(service)
    if snaked_service in b_e_d:
        for d_k, d_v in b_e_d[snaked_service].items():
            byc.update({d_k: d_v})

    # update response_entity_id from path
    update_entity_ids_from_path(byc)

    # update response_entity_id from form
    update_requested_schema_from_request(byc)
    set_response_entity(byc)

    set_io_params(byc)
    # CAVE: At this time the genome is just taken from the default so has to 
    # be run again after form value parsing ...
    generate_genomic_mappings(byc)
    set_special_modes(byc)

    # TODO: standardize the general defaults / entity defaults / form values merging
    #       through pre-parsing into identical structures and then use deepmerge etc.


################################################################################

def set_special_modes(byc):
    form = byc["form_data"]
    for m in ["test_mode", "debug_mode", "download_mode", "include_handovers"]:
        if m in form:
            v = test_truthy(form.get(m, False))
            byc.update({m: v})

    t_m_k = form.get("test_mode_count", "___none___")
    if re.match(r'^\d+$', str(t_m_k)):
        byc.update({"test_mode_count": int(t_m_k)})


################################################################################

def set_io_params(byc):
    form = byc["form_data"]

    if not "pagination" in byc:
        byc.update({"pagination": {"skip": 0, "limit": 0}})

    for sp in ["skip", "limit"]:
        if sp in form:
            if re.match(r'^\d+$', str(form[sp])):
                s_v = int(form[sp])
                byc["pagination"].update({sp: s_v})

    byc.update({"output": form.get("output", byc["output"])})

    # TODO: this is only used in some services ...
    if "method_keys" in byc["service_config"]:
        m = form.get("method", "___none___")
        if m in byc["service_config"]["method_keys"].keys():
            byc["method"] = m


################################################################################

def update_entity_ids_from_path(byc):
    if not byc["request_entity_path_id"]:
        return

    if not byc["response_entity_path_id"]:
        byc.update({"response_entity_path_id": byc["request_entity_path_id"]})

    req_p_id = byc["request_entity_path_id"]
    res_p_id = byc["response_entity_path_id"]

    m_k = byc["request_path_root"] + "_mappings"
    b_mps = byc[m_k]
    p_r_m = b_mps["path_response_type_mappings"]

    byc.update({
        "request_entity_id": p_r_m.get(req_p_id, byc["request_entity_path_id"]),
        "response_entity_id": p_r_m.get(res_p_id, byc["response_entity_path_id"])
    })


################################################################################

def update_requested_schema_from_request(byc):
    m_k = byc["request_path_root"] + "_mappings"
    b_mps = byc[m_k]
    form = byc["form_data"]
    b_qm = byc["query_meta"]

    if "requested_schema" in form:
        byc.update({"response_entity_id": form.get("requested_schema", byc["response_entity_id"])})
    elif "requested_schemas" in b_qm:
        byc.update({"response_entity_id": b_qm["requested_schemas"][0].get("entity_type", byc["response_entity_id"])})


################################################################################

def set_response_entity(byc):
    r_p_r = byc.get("request_path_root", "beacon")
    s_k = f'{r_p_r}_entry_types'
    b_rt_s = byc.get(s_k, {})
    r_e_id = byc.get("response_entity_id", "___none___")

    if r_e_id not in b_rt_s.keys():
        return

    byc.update({"response_entity": b_rt_s.get(r_e_id)})


################################################################################

def run_result_sets_beacon(byc):
    sr_r = byc["service_response"]["response"]

    byc.update({"dataset_results": {}})

    ######## result sets loop ##################################################

    for i, r_set in enumerate(sr_r["result_sets"]):

        ds_id = r_set["id"]
        generate_dataset_queries(byc, ds_id)

        r_set, r_s_res = populate_result_set(r_set, byc)

        if r_set["results_count"] < 1:
            continue

        # TODO: This condition avoids the "range on range" if previously set for
        # a handover query (by accessioinId)
        if not "range" in byc["pagination"]:
            set_pagination_range(len(r_s_res), byc)
        if test_truthy(byc["form_data"].get("paginate_results", True)):
            r_s_res = paginate_list(r_s_res, byc)

        check_alternative_single_set_deliveries(ds_id, r_s_res, byc)
        r_s_res = reshape_resultset_results(ds_id, r_s_res, byc)

        sr_r["result_sets"][i].update({
            "paginated_results_count": len(r_s_res),
            "exists": True,
            "results": r_s_res
        })

    ######## end of result sets loop ###########################################

    sr_rs = byc["service_response"].get("response_summary", {})
    sr_rs.update({"num_total_results": 0})
    for r_set in sr_r["result_sets"]:
        sr_rs["num_total_results"] += r_set["results_count"]

    if sr_rs["num_total_results"] > 0:
        sr_rs.update({"exists": True})


################################################################################

def populate_result_set(r_set, byc):
    ds_id = r_set["id"]
    execute_bycon_queries(ds_id, byc)

    # # Special check-out here since this forces a single handover
    # check_plot_responses(ds_id, byc)

    r_set.update({"results_handovers": dataset_response_add_handovers(ds_id, byc)})
    r_s_res = retrieve_data(ds_id, byc)

    r_set_update_counts(r_set, r_s_res, byc)

    return r_set, r_s_res


################################################################################

def r_set_update_counts(r_set, r_s_res, byc):
    ds_id = r_set["id"]
    ds_res = byc["dataset_results"][ds_id]

    r_set.update({"results_count": 0})
    counted = byc["config"].get("beacon_count_items", {})

    for c, c_d in counted.items():
        h_o_k = c_d.get("h->o_key", None)
        if h_o_k is None:
            continue
        if h_o_k not in ds_res:
            continue
        r_c = ds_res[h_o_k]["target_count"]
        s_c = len(list(set(ds_res[h_o_k]["target_values"])))
        r_set["info"]["counts"].update({c: r_c})

    if byc["empty_query_all_count"]:
        r_set.update({"results_count": byc["empty_query_all_count"]})
    elif isinstance(r_s_res, list):
        r_set.update({"results_count": len(r_s_res)})

    return r_set


################################################################################

def check_alternative_single_set_deliveries(ds_id, r_s_res, byc):
    check_datatable_delivery(r_s_res, byc)
    check_alternative_variant_deliveries(ds_id, byc)
    check_alternative_callset_deliveries(ds_id, byc)

    return


################################################################################

def response_meta_add_request_summary(r, byc):
    if not "received_request_summary" in r["meta"]:
        return r

    r_rcvd_rs = r["meta"]["received_request_summary"]
    defs = byc.get("beacon_defaults", {})
    b_e_d = defs.get("entity_defaults", {"info":{}})

    form = byc["form_data"]

    r_rcvd_rs.update({
        "filters": byc.get("filters", []),
        "pagination": byc.get("pagination", {}),
        "api_version": b_e_d["info"].get("api_version", "v2")
    })

    for p in ["include_resultset_responses", "requested_granularity"]:
        if p in form:
            r_rcvd_rs.update({p: form.get(p)})
            if "requested_granularity" in p:
                r["meta"].update({"returned_granularity": form.get(p)})

    try:
        for rrs_k, rrs_v in byc["service_config"]["meta"]["received_request_summary"].items():
            r_rcvd_rs.update({rrs_k: rrs_v})
    except:
        pass

    return r


################################################################################

def update_meta_queries(byc):
    try:
        if not "info" in byc["meta"]:
            byc["meta"].update({"info": {}})
        byc["service_response"]["meta"]["info"].update({"processed_query": byc.get("original_queries", {})})
    except:
        pass


################################################################################

def create_empty_beacon_response(byc):
    r_s = byc["response_entity"].get("response_schema", None)
    if r_s is None:
        return

    r, e = instantiate_response_and_error(byc, r_s)

    try:
        r["response"].update({"result_sets": []})
    except KeyError:
        pass

    if "beaconResultsetsResponse" in r_s:

        # TODO: stringent definition on when this is being used
        r_set = object_instance_from_schema_name(byc, "beaconResultsets", "definitions/ResultsetInstance")

        if "dataset_ids" in byc:
            for ds_id in byc["dataset_ids"]:
                ds_rset = r_set.copy()
                ds_rset.update({
                    "id": ds_id,
                    "set_type": "dataset",
                    "results_count": 0,
                    "exists": False,
                    "info": {"counts": {}}
                })
                r["response"]["result_sets"].append(ds_rset)

    byc.update({"service_response": r, "error_response": e})


################################################################################

def create_empty_service_response(byc):
    r, e = instantiate_response_and_error(byc, byc["response_entity"]["response_schema"])
    byc.update({"service_response": r, "error_response": e})


###############################################################################

def create_empty_non_data_response(byc):
    r, e = instantiate_response_and_error(byc, byc["response_entity"]["response_schema"])
    response_update_meta(r, byc)

    for r_k, r_v in byc["service_config"]["response"].items():
        r["response"].update({r_k: r_v})

    byc.update({"service_response": r, "error_response": e})


################################################################################

def check_plot_responses(ds_id, byc):
    if not "plot" in byc["output"]:
        return

    check_histoplot_plot_response(ds_id, byc)


################################################################################

def check_histoplot_plot_response(ds_id, byc):
    h_o_types = byc["handover_definitions"]["h->o_types"]
    ds_h_o = byc["dataset_definitions"][ds_id].get("handoverTypes", h_o_types.keys())

    if "histoplot" not in ds_h_o:
        return

    byc["include_handovers"] = True
    byc["dataset_definitions"][ds_id].update({"handoverTypes": ["histoplot"]})
    dataset_results_save_handovers(ds_id, byc)
    r_s_ho = dataset_response_add_handovers(ds_id, byc)

    for h_o in r_s_ho:
        if "histoplot" in h_o["handover_type"]["id"]:
            print_uri_rewrite_response(h_o["url"], "")
            exit()


################################################################################

def check_datatable_delivery(results, byc):
    if not "table" in byc["output"]:
        return
    if not "datatable_mappings" in byc:
        return

    export_datatable_download(results, byc)


################################################################################

def instantiate_response_and_error(byc, schema):
    """The response relies on the pre-processing of input parameters (queries etc)."""
    r = object_instance_from_schema_name(byc, schema, "")
    m = object_instance_from_schema_name(byc, "beaconResponseMeta", "")
    r.update({"meta": m})
    e = object_instance_from_schema_name(byc, "beaconErrorResponse", "")
    response_update_meta(r, byc)
    error_response_set_defaults(e)

    return r, e


################################################################################

def response_update_meta(r, byc):
    if "response_summary" in r:
        r_s = r.get("response_summary")
        if r_s is None:
            r_s = {"exists": False}
        else:
            r_s.update({"exists": False})
        r.update({"response_summary": r_s})
    response_meta_set_info_defaults(r, byc)
    response_meta_set_config_defaults(r, byc)
    response_meta_set_entity_values(r, byc)
    response_meta_add_request_summary(r, byc)
    r["meta"].update({"test_mode": byc.get("test_mode", False)})


################################################################################

def error_response_set_defaults(e):
    e.update({
        "error" : {
            "error_code": 200,
            "error_message": ""
        }
    })


################################################################################

def response_meta_set_info_defaults(r, byc):
    defs = byc.get("beacon_defaults", {})
    b_e_d = defs.get("entity_defaults", {"info":{}})

    # TODO: command line hack ...
    for i_k in ["api_version", "beacon_id"]:
        if "meta" in r and "info" in b_e_d:
            r["meta"].update({i_k: b_e_d["info"].get(i_k, "")})


################################################################################

def response_meta_set_config_defaults(r, byc):
    # TODO: should be in schemas
    m = byc["service_config"].get("meta", {})
    for k, v in m.items():
        r["meta"].update({k: v})


################################################################################

def response_meta_set_entity_values(r, byc):
    try:
        r["meta"]["received_request_summary"].update({
            "requested_schemas": [byc["response_entity"]["beacon_schema"]]
        })
        r["meta"].update({"returned_schemas": [byc["response_entity"]["beacon_schema"]]})
    except:
        pass


################################################################################

def response_add_received_request_summary_parameters(byc):
    if not "received_request_summary" in byc["service_response"].get("meta", {}):
        return

    for name in ["method", "dataset_ids", "filters", "variant_pars", "test_mode"]:
        value = byc.get(name, False)
        if value is False:
            continue
        if "variant_pars" in name:
            name = "request_parameters"
        byc["service_response"]["meta"]["received_request_summary"].update({name: value})

################################################################################

def received_request_summary_add_custom_parameter(byc, parameter, value):
    if not "received_request_summary" in byc["service_response"].get("meta", {}):
        return

    byc["service_response"]["meta"]["received_request_summary"].update({parameter: value})


################################################################################

def response_collect_errors(byc):
    # TODO: flexible list of errors
    if not byc["queries"].keys():
        response_add_error(byc, 200, "No (correct) query parameters were provided.")
    if len(byc["dataset_ids"]) > 1:
        response_add_error(byc, 200, "More than 1 `datasetIds` value was provided.")


################################################################################

def response_add_error(byc, code=200, message=False):
    if message is False:
        return
    if len(str(message)) < 1:
        return

    e = {"error_code": code, "error_message": message}
    byc["error_response"].update({"error": e})


################################################################################

def response_add_warnings(byc, message=False):
    if message is False:
        return
    if len(str(message)) < 1:
        return

    if not "service_response" in byc:
        return

    if not "info" in byc["service_response"]:
        byc["service_response"].update({"info": {}})
    if not "warnings" in byc["service_response"]:
        byc["service_response"]["info"].update({"warnings": []})

    byc["service_response"]["info"]["warnings"].append(message)


################################################################################

def collations_set_delivery_keys(byc):
    # the method keys can be overriden with "deliveryKeys"

    method = byc.get("method", False)
    conf = byc["service_config"]

    d_k = form_return_listvalue(byc["form_data"], "deliveryKeys")

    if len(d_k) > 0:
        return d_k

    if method is False:
        return d_k

    # here the empty list is returned, leading to all
    if "details" in method:
        return d_k

    if not "method_keys" in conf:
        return d_k

    if method in conf["method_keys"]:
        d_k = conf["method_keys"][method]

    return d_k


################################################################################

def populate_service_response(byc, results):
    byc["service_response"].update({
        "response_summary": {
            "exists": False,
            "num_total_results": 0
        }
    })

    populate_service_header(byc, results)
    populate_service_response_counts(byc)

    if len(results) > 0:
        byc["service_response"]["response_summary"].update({
            "exists": True,
            "num_total_results": len(results)
        })

    if byc["service_response"]["response"] is None:
        byc["service_response"].update({"response": {"results":results}})
    else:
        byc["service_response"]["response"].update({"results": results})

    check_switch_to_count_response(byc)
    check_switch_to_boolean_response(byc)


################################################################################

def populate_service_header(byc, results):
    try:
        r_s = byc["service_response"]["response_summary"]
    except:
        return

    r_no = 0

    if isinstance(results, list):
        r_no = len(results)
        r_s.update({"num_total_results": r_no})
    if r_no > 0:
        r_s.update({"exists": True})


################################################################################

def populate_service_response_counts(byc):
    if not "dataset_results" in byc:
        return
    if not "dataset_ids" in byc:
        return

    ds_id = byc["dataset_ids"][0]

    if not ds_id in byc["dataset_results"]:
        return

    counts = {}
    counted = byc["config"].get("beacon_count_items", {})

    for c, c_d in counted.items():
        h_o_k = c_d.get("h->o_key")
        if not h_o_k:
            continue
        if h_o_k in byc["dataset_results"][ds_id]:
            counts[c] = byc["dataset_results"][ds_id][h_o_k]["target_count"]

    byc["service_response"]["info"].update({"counts": counts})


################################################################################

def check_alternative_variant_deliveries(ds_id, byc):
    if not "genomicVariant" in byc["response_entity_id"]:
        return

    if "pgxseg" in byc["output"]:
        export_pgxseg_download(ds_id, byc)

    if "variants" in byc["method"]:
        export_variants_download(ds_id, byc)


################################################################################

def check_alternative_callset_deliveries(ds_id, byc):
    check_callsets_matrix_delivery(ds_id, byc)


################################################################################

def return_filtering_terms_response(byc):
    if not "filteringTerm" in byc["response_entity_id"]:
        return

    # TODO: correct response w/o need to fix
    # TODO: dataset specificity etc.
    byc["service_response"].update({"response": {"filteringTerms": [], "resources": []}})

    f_r_d = {}

    f_coll = byc["config"]["filtering_terms_coll"]

    f_t_s = []
    ft_fs = []

    if "filters" in byc:
        if len(byc["filters"]) > 0:
            for f in byc["filters"]:
                ft_fs.append('(' + f["id"] + ')')
    f_s = '|'.join(ft_fs)
    f_re = re.compile(r'^' + f_s)

    collation_types = set()

    scopes = ["biosamples", "individuals", "analyses", "genomicVariations"]

    for ds_id in byc["dataset_ids"]:

        query = {}

        try:
            if len(byc["form_data"]["scope"]) > 4:
                query.update({"scope": byc["form_data"]["scope"]})
        except:
            pass

        fields = {"_id": 0}

        f_s, e = mongo_result_list(ds_id, f_coll, query, fields)

        t_f_t_s = []

        for f in f_s:
            collation_types.add(f.get("collation_type", None))
            f_t = {"count": f.get("count", 0)}
            for k in ["id", "type", "label"]:
                if k in f:
                    f_t.update({k: f[k]})
            f_t.update({"scopes": scopes})
            t_f_t_s.append(f_t)

        f_t_s.extend(t_f_t_s)

    byc["service_response"]["response"].update({"filteringTerms": f_t_s})
    byc["service_response"]["response"].update({"resources": create_filters_resource_response(collation_types, byc)})
    byc["service_response"]["response_summary"].update({"num_total_results": len(f_t_s)})
    if len(f_t_s) > 0:
        byc["service_response"]["response_summary"].update({"exists": True})

    cgi_print_response(byc, 200)


################################################################################

def create_filters_resource_response(collation_types, byc):
    r_o = {}
    resources = []

    f_d_s = byc["filter_definitions"]
    collation_types = list(collation_types)
    res_schema = object_instance_from_schema_name(byc, "beaconFilteringTermsResults", "definitions/Resource",
                                                  "json")
    for c_t in collation_types:
        f_d = f_d_s[c_t]
        r = {}
        for k in res_schema.keys():
            if k in f_d:
                r.update({k: f_d[k]})

        r_o.update({f_d["namespace_prefix"]: r})

    for k, v in r_o.items():
        resources.append(v)
    return resources


################################################################################

def check_callset_plot_delivery(byc):
    if not "samplesplot" in byc["output"]:
        return

    results = []
    p_r = byc["pagination"]

    for ds_id in byc["dataset_results"].keys():

        ds_results = byc["dataset_results"][ds_id]
        if not "callsets._id" in ds_results:
            continue

        v_d = byc["variant_parameters"]
        mongo_client = MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
        cs_coll = mongo_client[ds_id]["callsets"]
        var_coll = mongo_client[ds_id]["variants"]

        cs_r = ds_results["callsets._id"]
        if len(cs_r) < 1:
            continue

        q_vals = cs_r["target_values"]
        r_no = len(q_vals)
        if r_no > p_r["limit"]:
            q_vals = paginate_list(q_vals, byc)

        for cs in cs_coll.find({"_id": {"$in": q_vals}}):

            cs_id = cs.get("id", "NA")

            cnv_chro_stats = cs.get("cnv_chro_stats", False)
            cnv_statusmaps = cs.get("cnv_statusmaps", False)

            if cnv_chro_stats is False or cnv_statusmaps is False:
                continue

            p_o = {
                "dataset_id": ds_id,
                "callset_id": cs_id,
                "biosample_id": cs.get("biosample_id", "NA"),
                "cnv_chro_stats": cs.get("cnv_chro_stats", {}),
                "cnv_statusmaps": cs.get("cnv_statusmaps", {}),
                "probefile": callset_guess_probefile_path(cs, byc),
                "variants": []
            }

            if r_no == 1 and p_o["probefile"] is not False:
                p_o.update({"cn_probes": ByconBundler(byc).read_probedata_file(p_o["probefile"]) })

            v_q = {"callset_id": cs_id}

            for v in var_coll.find(v_q):

                v_r = de_vrsify_variant(v, byc)
                if v is False:
                    return v

                p_o["variants"].append(v_r)

            results.append(p_o)

    plot_data_bundle = {"callsets_variants_bundles": results}
    ByconPlot(byc, plot_data_bundle).svg_response()


################################################################################

def check_computed_histoplot_delivery(byc):
    if not "histo" in byc.get("output", "___none___"):
        return

    f_d = byc.get("filter_definitions", {})

    p_r = byc["pagination"]
    interval_sets = []

    for ds_id in byc["dataset_results"].keys():

        ds_results = byc["dataset_results"][ds_id]

        if not "callsets._id" in ds_results:
            continue

        mongo_client = MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
        bios_coll = mongo_client[ds_id]["biosamples"]
        cs_coll = mongo_client[ds_id]["callsets"]

        f_s_dists = []
        f_s_k = ""
        f_s_t = byc["form_data"].get("plot_group_by", "___none___")

        if f_s_t in f_d.keys():

            if not "biosamples._id" in ds_results:
                continue

            bios_q_v = ds_results["biosamples._id"].get("target_values", [])
            if len(bios_q_v) < 1:
                continue

            f_s_k = f_d[f_s_t].get("db_key", "___none___")
            f_s_p = f_d[f_s_t].get("pattern", False)
            f_s_q = {"_id": {"$in": bios_q_v}}
            f_s_dists = bios_coll.distinct(f_s_k, f_s_q)
            if f_s_p is not False:
                r = re.compile(f_s_p)
                f_s_dists = list(filter(lambda d: r.match(d), f_s_dists))

            for f_s_id in f_s_dists:

                bios_id_q = {"$and": [
                    {f_s_k: f_s_id},
                    {"_id": {"$in": bios_q_v}}
                ]}

                bios_ids = bios_coll.distinct("id", bios_id_q)
                cs__ids = cs_coll.distinct("_id", {"biosample_id": {"$in": bios_ids}})
                r_no = len(cs__ids)
                if r_no > p_r["limit"]:
                    cs__ids = paginate_list(cs__ids, byc)

                label = f"Search Results (subset {f_s_id})"

                iset = callset__ids_create_iset(ds_id, label, cs__ids, byc)
                interval_sets.append(iset)

        else:
            cs_r = ds_results["callsets._id"]
            cs__ids = cs_r["target_values"]
            r_no = len(cs__ids)
            if r_no > p_r["limit"]:
                cs__ids = paginate_list(cs__ids, byc)

            iset = callset__ids_create_iset(ds_id, "Search Results", cs__ids, byc)
            interval_sets.append(iset)

    plot_data_bundle = {"interval_frequencies_bundles": interval_sets}
    ByconPlot(byc, plot_data_bundle).svg_response()


################################################################################

def check_computed_interval_frequency_delivery(byc):
    if not "frequencies" in byc["output"]:
        return

    ds_id = byc["dataset_ids"][0]
    ds_results = byc["dataset_results"][ds_id]
    p_r = byc["pagination"]

    if not "callsets._id" in ds_results:
        return

    cs_r = ds_results["callsets._id"]

    mongo_client = MongoClient(host=environ.get("BYCON_MONGO_HOST", "localhost"))
    cs_coll = mongo_client[ds_id]["callsets"]

    open_text_streaming(byc["env"], "interval_cnv_frequencies.pgxseg")

    for d in ["id", "assemblyId"]:
        print("#meta=>{}={}".format(d, byc["dataset_definitions"][ds_id][d]))
    print_filters_meta_line(byc)
    print('#meta=>query="{}"'.format(json.dumps(byc["queries_at_execution"], indent=None, sort_keys=True, default=str)))
    print('#meta=>skip={};limit={}'.format(p_r["skip"], p_r["limit"]))
    print("#meta=>sample_no={}".format(cs_r["target_count"]))

    q_vals = cs_r["target_values"]
    r_no = len(q_vals)
    if r_no > p_r["limit"]:
        q_vals = paginate_list(q_vals, byc)
        print(
            '#meta=>"WARNING: Only analyses {} - {} (out of {}) used for calculations due to pagination skip and limit"'.format(
                (p_r["range"][0] + 1), p_r["range"][-1], cs_r["target_count"]))

    h_ks = ["reference_name", "start", "end", "gain_frequency", "loss_frequency", "no"]
    print("group_id\t" + "\t".join(h_ks))

    cs_cursor = cs_coll.find({"_id": {"$in": q_vals}, "variant_class": {"$ne": "SNV"}})

    intervals, cnv_cs_count = interval_counts_from_callsets(cs_cursor, byc)

    for intv in intervals:
        v_line = []
        v_line.append("query_result")
        for k in h_ks:
            v_line.append(str(intv[k]))
        print("\t".join(v_line))

    close_text_streaming()


################################################################################

def check_callsets_matrix_delivery(ds_id, byc):
    if not "pgxmatrix" in byc["output"]:
        return

    export_callsets_matrix(ds_id, byc)


################################################################################

def bycon_bundle_create_intervalfrequencies_object(bycon_bundle, byc):
    return callsets_create_iset("import", "", bycon_bundle["callsets"], byc)

################################################################################



import argparse
import re

from humps import decamelize


################################################################################

def get_bycon_args(byc):
    if byc.get("check_args", False) is False:
        return

    # Serves as "we've been here before" marker - before the env check.
    byc.update({"check_args": False})

    if "local" not in byc["env"]:
        return

    if byc["script_args"]:
        for d_d in byc["argument_definitions"]["parameters"]:
            a_d_k = list(d_d["argDef"].keys())[0]
            if a_d_k not in byc["script_args"]:
                byc["argument_definitions"].pop(a_d_k, None)

    create_args_parser(byc)


################################################################################

def create_args_parser(byc):
    if not "local" in byc["env"]:
        return

    arg_defs = byc["argument_definitions"]["parameters"]

    parser = argparse.ArgumentParser()
    for d_d in arg_defs:
        a_d_k = list(d_d["argDef"].keys())[0]
        defs = d_d["argDef"][a_d_k]
        parser.add_argument(*defs.pop("flags"), **defs)
    byc.update({"args": parser.parse_args()})


################################################################################

def args_update_form(byc):
    if not "local" in byc["env"]:
        return

    arg_defs = byc["argument_definitions"]["parameters"]
    list_pars = []
    for d_d in arg_defs:
        a_d_k = list(d_d["argDef"].keys())[0]
        if "array" in d_d.get("type", "string"):
            list_pars.append(a_d_k)

    arg_vars = vars(byc["args"])

    for p in arg_vars.keys():
        if arg_vars[p] is None:
            continue
        p_d = decamelize(p)
        if p in list_pars:
            byc["form_data"].update({p_d: arg_vars[p].split(',')})
        else:
            byc["form_data"].update({p_d: arg_vars[p]})


################################################################################

def genome_binning_from_args(byc):
    if byc["args"].key:
        byc.update({"genome_binning": byc["args"].key})


################################################################################

def filters_from_args(byc):
    if not "args" in byc:
        return

    if not "filters" in byc:
        byc.update({"filters": []})

    if byc["args"].filters:
        for f in re.split(",", byc["args"].filters):
            byc["filters"].append({"id": f})


################################################################################

def set_collation_types(byc):
    if byc["args"].collationtypes:
        s_p = {}
        for p in re.split(",", byc["args"].collationtypes):
            if p in byc["filter_definitions"].keys():
                s_p.update({p: byc["filter_definitions"][p]})
        if len(s_p.keys()) < 1:
            print("No existing collation type was provided with -c ...")
            exit()

        byc.update({"filter_definitions": s_p})


################################################################################

def set_processing_modes(byc):
    byc.update({"update_mode": False})

    try:
        if byc["test_mode"] is True:
            byc.update({"update_mode": False})
            print("¡¡¡ TEST MODE - no db update !!!")
            return
    except:
        pass

    try:
        if byc["args"].update:
            byc.update({"update_mode": True})
            print("¡¡¡ UPDATE MODE - may overwrite entries !!!")
    except:
        pass
